/*
 * $QNXLicenseC:
 * Copyright 2014, QNX Software Systems.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You
 * may not reproduce, modify or distribute this software except in
 * compliance with the License. You may obtain a copy of the License
 * at: http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as
 * contributors under the License or as licensors under other terms.
 * Please review this entire file for other proprietary rights or license
 * notices, as well as the QNX Development Suite License Guide at
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */

/*
 * i.MX6Q Nitrogen board with Cortex-A9 MPCore
 */

#include "startup.h"
#include <time.h>
#include "board.h"

extern int dcache_enable;
extern int enable_TZAC;
#include <stdbool.h>

extern struct callout_rtn reboot_mx6x;
extern void mx6q_usb_otg_host_init(void);
extern void mx6q_usb_host1_init(void);
extern void mx6q_init_ecspi(void);
extern void mx6q_init_i2c1(void);
extern void mx6q_init_i2c2(void);
extern void mx6q_init_i2c3(void);
extern void mx6q_init_usdhc3(void);
extern void mx6q_init_usdhc4(void);
extern void mx6q_init_displays(void);
extern void mx6q_init_lvds(void);
extern void mx6q_init_lcd_panel(void);
extern void mx6q_init_can(void);
extern void init_audio(void);
extern void mx6q_init_audmux_pins(void);
extern void mx6q_init_enet_pre_reset(void);
extern void mx6q_init_enet_post_reset(void);
extern void mx6q_init_mac(void);
extern void mx6q_init_camera(void);
extern unsigned mx6x_init_lvds_clock(void);

const struct callout_slot callouts[] = {
	{ CALLOUT_SLOT( reboot, _mx6x) },
};

const struct debug_device debug_devices[] = {
	{	"mx1",
		{"0x021E8000^0.115200.80000000.16",
		},
		init_mx1,
		put_mx1,
		{	&display_char_mx1,
			&poll_key_mx1,
			&break_detect_mx1,
		}
	},
};


unsigned mx6x_per_clock = 0;
uint32_t uart_clock;


/*
 * main()
 *	Startup program executing out of RAM
 *
 * 1. It gathers information about the system and places it in a structure
 *    called the system page. The kernel references this structure to
 *    determine everything it needs to know about the system. This structure
 *    is also available to user programs (read only if protection is on)
 *    via _syspage->.
 *
 * 2. It (optionally) turns on the MMU and starts the next program
 *    in the image file system.
 */
int
main(int argc, char **argv, char **envv)
{
	int		opt, options = 0;
	bool		imx6_quad_dual;
	bool		imx6_quad_plus;

	/*
	/ Note: The last 256 MB of SDRAM is reserved for the GPU when TrustZone is enabled
	/ gpu_reserved_start: Start address of reserved SDRAM for GPU usage
	*/
	unsigned gpu_reserved_start = MX6X_SDRAM_64_BIT_BASE + MEG(MX6X_SDRAM_SIZE - GPU_SDRAM_SIZE);

	/*
	 * Initialise debugging output
	 */
	select_debug(debug_devices, sizeof(debug_devices));

	add_callout_array(callouts, sizeof(callouts));

	enable_TZAC = FALSE;

	/* identify the imx6 processor we're running on */
	imx6_quad_dual = (get_mx6_chip_type() == MX6_CHIP_TYPE_QUAD_OR_DUAL) ? TRUE : FALSE;
	imx6_quad_plus = (imx6_quad_dual && (get_mx6_chip_rev() == MX6_CHIP_REV_2_0)) ? TRUE : FALSE;

	// common options that should be avoided are:
	// "AD:F:f:I:i:K:M:N:o:P:R:S:Tvr:j:Z"
	while ((opt = getopt(argc, argv, COMMON_OPTIONS_STRING "mtW")) != -1) {
		switch (opt) {
		case 'm':
			dcache_enable = TRUE;
			break;
		case 't':
			enable_TZAC = TRUE;
			break;
		case 'W':
			options |= MX6X_WDOG_ENABLE;
			break;
		default:
			handle_common_option(opt);
			break;
		}
	}

	if (options &  MX6X_WDOG_ENABLE) {
		/*
		 * Enable WDT
		 */
		mx6x_wdg_reload();
		mx6x_wdg_enable();
	}

	/* Make sure to not allocate anything in the SDRAM which is in GPU SDRAM TrustZone Area */
	if (enable_TZAC)
		avoid_ram(gpu_reserved_start, MEG(GPU_SDRAM_SIZE));

	/*
	 * Collect information on all free RAM in the system
	 */
	init_ram(gpu_reserved_start, MX6X_SDRAM_SIZE);

	/*
	 * set CPU frequency, currently max stable CPU freq is 792MHz
	 */
	if (cpu_freq == 0)
		cpu_freq = mx6x_get_cpu_clk();

	/*
	 * Remove RAM used by modules in the image
	 */
	alloc_ram(shdr->ram_paddr, shdr->ram_size, 1);

	/*
	 * Initialize SMP
	 */
	init_smp();

	/*
	 * Initialize MMU
	 */
	if (shdr->flags1 & STARTUP_HDR_FLAGS1_VIRTUAL)
		init_mmu();

	/* Initialize the Interrupts related Information */
	init_intrinfo();

	/* Initialize the Timer related information */
	mx6x_init_qtime();

	/* Init Clocks (must happen after timer is initialized */
	mx6x_init_clocks();

	/* Init L2 Cache Controller */
	init_cacheattr();

	/* Initialize the CPU related information */
	init_cpuinfo();

	/* Initialize the Hwinfo section of the Syspage */
	init_hwinfo();

	if(imx6_quad_dual) {
		if(imx6_quad_plus)
			add_typed_string(_CS_MACHINE, "i.MX6 QuadPlus Nitrogen Board");
		else
			add_typed_string(_CS_MACHINE, "i.MX6 Quad Nitrogen Board");
	}

	/* AIPSTZ init */
	mx6x_init_aipstz();

	/* ENET */
	mx6q_init_mac();
	mx6q_init_enet_pre_reset();

#ifndef SABRELITE
	/* Pull GPIO1[27] down to reset the PHY */
	mx6x_reset_gpio_pin(MX6X_GPIO1_BASE, 27, 0);
#else
	/* Pull GPIO3[23] down to reset the PHY */
	mx6x_reset_gpio_pin(MX6X_GPIO3_BASE, 23, 0);
#endif

	/* Init CAN Controller */
	mx6q_init_can();

	/* Configure PIN MUX to enable i2c */
	mx6q_init_i2c1();
	mx6q_init_i2c2();
	mx6q_init_i2c3();

	/* Init audio to support SGTL5000 codec */
	mx6q_init_audmux_pins();
	init_audio();

	/* Test if LDO_PU is gated (responsible for VPU/GPU power domain) */
	if (pmu_get_voltage(LDO_PU) == CPU_POWER_GATED_OFF_VOLTAGE)
	{
		pmu_power_up_gpu();
	}

	/* Make sure to reset IPU IDMAC before display setup */
	out32(MX6X_IPU1_BASE, in32(MX6X_IPU1_BASE) | 1 << 22);
	out32(MX6X_IPU1_BASE, in32(MX6X_IPU1_BASE) & ~(1 << 22));

	/* Set GPU3D clocks */
	mx6x_init_gpu3D();

	/*
	 * If using an LVDS display mx6x_init_lvds_clock() should be called which will select
	 * PLL5 as the LDB clock source.
	 */
	mx6x_init_lvds_clock();

	/* Configure PIN MUX to enable SD */
	mx6q_init_usdhc3();
	mx6q_init_usdhc4();

	/* Init USB OTG, H1 */
	mx6q_usb_otg_host_init();
	mx6q_usb_host1_init();

	/* Configure PIN MUX to enable SPI */
	mx6q_init_ecspi();

	/* Configure pins for LVDS, LCD display*/
	mx6q_init_displays();

	/* SATA */
	if (get_mx6_chip_type() == MX6_CHIP_TYPE_QUAD_OR_DUAL) {
		mx6x_init_sata(ANATOP_PLL8_ENET_REF_ENET_50M);
	}

	/* PHY requires 10ms reset time */
	mx6x_reset_gpio_pin_fin(10000);
	/* Strap hold time for AR8035 */
	mx6x_usleep(12);
	mx6q_init_enet_post_reset();

	mx6x_dump_clocks();

	mx6q_init_camera();

	/*
	 * Load bootstrap executables in the image file system and Initialise
	 * various syspage pointers. This must be the _last_ initialisation done
	 * before transferring control to the next program.
	 */
	init_system_private();

	/*
	 * This is handy for debugging a new version of the startup program.
	 * Commenting this line out will save a great deal of code.
	 */
	print_syspage();
	return 0;
}
